--[[============================================================================
envelope_processing.lua
============================================================================]]--


function configure_envelope_loop()
  --Sets the loop properties of each envelope involved.
  --Also does several sanity checks
  local song = renoise.song()
  local sel_ins = song.selected_instrument_index
  local ins = renoise.song().instruments[sel_ins].sample_envelopes
  if note_loop_end == nil then
     return
  end
  
  if note_loop_end + 1 > ins.pitch.length then
    ins.pitch.length = note_loop_end +1
  end
  if note_loop_start == nil then
    return
  end
  if note_loop_start + 1 > ins.pitch.length then
    ins.pitch.length  = note_loop_start + 1
  end
  if vol_loop_end == nil then
     return
  end
  if envelope_volume_toggle then
    if vol_loop_end + 1 > ins.volume.length then
      ins.volume.length = vol_loop_end +1
    end
    if vol_loop_start == nil then
      return
    end
    if vol_loop_start + 1 > ins.volume.length then
      ins.volume.length  = vol_loop_start + 1
    end
  end
  if pan_loop_end == nil then
     return
  end
  if envelope_panning_toggle then
    if pan_loop_end + 1 > ins.pan.length then
      ins.pan.length = pan_loop_end +1
    end
    if pan_loop_start == nil then
      return
    end
    if pan_loop_start + 1 > ins.pan.length then
      ins.pan.length  = pan_loop_start + 1
    end
  end
  
--  if env_auto_apply then

      ins.pitch.loop_mode = note_loop_type

      if note_loop_start < MAXIMUM_FRAME_LENGTH and note_loop_start > -1 then
        if note_loop_start +1 > ins.pitch.length then 
          ins.pitch.length = note_loop_start + 1
        end
        ins.pitch.loop_start = note_loop_start +1
      end

      if note_loop_end < MAXIMUM_FRAME_LENGTH and note_loop_end > -1 then
        if note_loop_end +1  > ins.pitch.length then 
          ins.pitch.length = note_loop_end + 1
        end
        ins.pitch.loop_end = note_loop_end +1
      end
    
    if envelope_volume_toggle then

        ins.volume.loop_mode = vol_loop_type
  
        if vol_loop_start < MAXIMUM_FRAME_LENGTH and vol_loop_start > -1 then
          if vol_loop_start +1 > ins.volume.length then 
            ins.volume.length = vol_loop_start + 1
          end
          ins.volume.loop_start = vol_loop_start +1
        end

        if vol_loop_end < MAXIMUM_FRAME_LENGTH and vol_loop_end > -1 then
          if vol_loop_end +1 > ins.volume.length then 
            ins.volume.length = vol_loop_end + 1
          end
          ins.volume.loop_end = vol_loop_end +1
        end
    end
    
    if envelope_panning_toggle then
        ins.pan.loop_mode = pan_loop_type
  
        if pan_loop_start < MAXIMUM_FRAME_LENGTH and pan_loop_start > -1 then
          if pan_loop_start +1 > ins.pan.length then 
            ins.pan.length = pan_loop_start + 1
          end
          ins.pan.loop_start = pan_loop_start +1
        end
  
        if pan_loop_end < MAXIMUM_FRAME_LENGTH and pan_loop_end > -1 then
          if pan_loop_end +1 > ins.pan.length then 
            ins.pan.length = pan_loop_end + 1
          end
          ins.pan.loop_end = pan_loop_end +1
        end
    end

--  end
  
  if construct then
    construct_envelope()
  end
  
end


---------------------------------------------------------------------------------------


function set_pitch_table()
--Tone scope output corrector
--Check the values if they are crossing boundaries (e.g.:[14] or [-15])
--and then set the values for the pitch frame to the maximum value
--according to the current tone factor. This way we have a safe zone without
--loosing the original values when crossing the boundaries.
  local song = renoise.song()
  local tool_instrument = song.selected_instrument_index
  local samples = #song.instruments[tool_instrument].samples
  local factor_selection = 0
  local bound_break = false
  local lowest_note = 0
  local highest_note = 0
  local octave_shift = 0
  
  if note_scheme_size == 0 or note_scheme_size == -1 then
    return
  end
  
--[[
  if env_note_value[note_scheme_size] ~= NOTE_SCHEME_TERMINATION then
    note_scheme_size = MAXIMUM_FRAME_LENGTH
  end
--]]
  for x = 0,note_scheme_size do
    pitch_table[x] = env_note_value[x]
    if tonumber(env_note_value[x]) == NOTE_SCHEME_TERMINATION then
      note_scheme_size = x
      break
    end
  end
  

  for _ = 0, note_scheme_size do
    if string.sub(pitch_table[_],1,1) == '[' then
      bound_break = true
      break
    end
  end

  if bound_break then
    for _ = 0, note_scheme_size do
      if string.sub(pitch_table[_],1,1) == '[' then
        local temp_pitch = pitch_table[_]
        temp_pitch = string.sub(temp_pitch,2,string.len(temp_pitch)-1)
        if tonumber(temp_pitch) < NOTE_SCHEME_TERMINATION then
        
          if string.sub(temp_pitch,1,1) == '-' then
            if tonumber(temp_pitch) < lowest_note then 
              lowest_note = tonumber(temp_pitch)
            end
          else
            if tonumber(temp_pitch) > highest_note then 
              highest_note = tonumber(temp_pitch)
            end
          end
        end
      end
    end
    
    if highest_note - lowest_note < 24 then
      octave_shift = 12*math.floor(highest_note/12)
      for _ = 0, note_scheme_size do
        local temp_pitch = pitch_table[_]
        if string.sub(pitch_table[_],1,1) == '[' then
          temp_pitch = string.sub(temp_pitch,2,string.len(temp_pitch)-1)
        end
        if tonumber(temp_pitch) < NOTE_SCHEME_TERMINATION then
          pitch_table[_] = tostring(tonumber(temp_pitch) - octave_shift)
        end
      end
    end
  end

  for _ = 0, note_scheme_size do
    if string.sub(pitch_table[_],1,1) == '[' then
      local temp_pitch = pitch_table[_]
      temp_pitch = string.sub(temp_pitch,2,string.len(temp_pitch)-1)
      if tonumber(temp_pitch) < NOTE_SCHEME_TERMINATION then
        if string.sub(temp_pitch,1,1) == '-' then
          if string.len(temp_pitch) == 3 then
            temp_pitch = '-12'
          end
          if string.len(temp_pitch) == 4 then
            temp_pitch = '-120'
          end
          if string.len(temp_pitch) > 4 then
            temp_pitch = '-1200'
          end
        else
          if string.len(temp_pitch) == 2 then
            temp_pitch = '12'
          end
          if string.len(temp_pitch) == 3 then
            temp_pitch = '120'
          end
          if string.len(temp_pitch) > 3 then
            temp_pitch = '1200'
          end
        end
        pitch_table[_] = tonumber(temp_pitch)
      end
    end

    if tonumber(pitch_table[_]) * tone_factor <= 1200 and tonumber(pitch_table[_]) * tone_factor >=  -1200 then
      pitch_table[_] = tonumber(pitch_table[_]) * tone_factor
    end

  end
  
  if construct then
    construct_envelope()
  end
  
end


---------------------------------------------------------------------------------------




function construct_envelope ()
  -- build all the envelopes and loops.
  local song = renoise.song()
  local tool_instrument = song.selected_instrument_index
  local instrument = song.instruments[tool_instrument].sample_envelopes

  
--  if env_auto_apply then
    local loop_start = note_loop_start
    local loop_end = note_loop_end
    local loop_type = note_loop_type
    local point_table = {}
    local loop_end_store = instrument.pitch.loop_end
   
    change_from_tool = true
    instrument.pitch.play_mode = PLAY_MODE_POINTS
    instrument.pitch.enabled = envelope_pitch_toggle
    
    instrument.pitch:clear_points()
    instrument.pitch.play_mode = env_pitch_type
   
    for x = 0,note_scheme_size do
      if pitch_table[x] ~= nil and tonumber(pitch_table[x]) ~= nil then --Notes outside the pitch scope fuck up again!!
        if tonumber(pitch_table[x]) < NOTE_SCHEME_TERMINATION then
          local ptone = 0.5 + (tonumber(pitch_table[x]) * TONE)
          instrument.pitch:add_point_at(x+1,ptone)
--          print('<point>'..tostring(x*256),','..tostring(ptone)..'<point>')
        end
      end
    end      

    if note_scheme_size < MINIMUM_FRAME_LENGTH then
      note_scheme_size = MINIMUM_FRAME_LENGTH
    end

    if note_scheme_size < MAXIMUM_FRAME_LENGTH then
      instrument.pitch.length = note_scheme_size + 1
    else
      instrument.pitch.length = MAXIMUM_FRAME_LENGTH
    end
    
    if note_sustain > -1 then
      if instrument.pitch.length >= note_sustain + 1 then
        instrument.pitch.sustain_position = note_sustain + 1
      else
        instrument.pitch.sustain_position = 1
        note_sustain = 1
      end
      instrument.pitch.sustain_enabled = true
    else
      instrument.pitch.sustain_enabled = false
    end

    if envelope_volume_toggle then
      change_from_tool = true
      instrument.volume.enabled = true
      instrument.volume:clear_points()
      instrument.volume.play_mode = env_volume_type
      for x = 0,vol_scheme_size do
        if env_vol_value[x] ~= nil then
          if tonumber(env_vol_value[x]) < VOL_PAN_TERMINATION then
            local pvol = (tonumber(env_vol_value[x]) / 100)
            instrument.volume:add_point_at(x+1,pvol)
          end

        end
      end      
      if vol_release > -1 then
        instrument.volume.fade_amount = vol_release
      end 
      if vol_sustain > -1 then
        if instrument.volume.length >= vol_sustain + 1 then
          instrument.volume.sustain_position = vol_sustain + 1
        else
          instrument.volume.sustain_position = 1
          vol_sustain = 1
        end
        instrument.volume.sustain_enabled = true
      else
        instrument.volume.sustain_enabled = false
      end
      if vol_scheme_size < MINIMUM_FRAME_LENGTH then
        vol_scheme_size = MINIMUM_FRAME_LENGTH
      end

      if vol_scheme_size < MAXIMUM_FRAME_LENGTH then
        instrument.volume.length = vol_scheme_size + 1
      else
        instrument.volume.length = MAXIMUM_FRAME_LENGTH
      end
    else
      instrument.volume.enabled = false
    end
    
    if envelope_panning_toggle then
      change_from_tool = true
      instrument.pan.enabled = true
      instrument.pan:clear_points()

      instrument.pan.play_mode = env_panning_type

      for x = 0,pan_scheme_size do
        if env_pan_value[x] ~= nil then
          if tonumber(env_pan_value[x]) < VOL_PAN_TERMINATION then
            local ppan = (tonumber(env_pan_value[x]) / 100)
            if ppan <= 0 then
              ppan = 0.5 - math.abs(ppan)
            else
              ppan = ppan + 0.5
            end
            instrument.pan:add_point_at(x+1,ppan)
          end
        end
      end      

      if pan_sustain > -1 then
        if instrument.pan.length >= pan_sustain + 1 then
          instrument.pan.sustain_position = pan_sustain + 1
        else
          instrument.pan.sustain_position = 1
          pan_sustain =1
        end
        
        instrument.pan.sustain_enabled = true
      else
        instrument.pan.sustain_enabled = false
        
      end

      if pan_scheme_size < MINIMUM_FRAME_LENGTH then
        pan_scheme_size = MINIMUM_FRAME_LENGTH
      end
     
      if pan_scheme_size < MAXIMUM_FRAME_LENGTH then
        instrument.pan.length = pan_scheme_size + 1
      else
        instrument.pan.length = MAXIMUM_FRAME_LENGTH
      end
    else
      instrument.pan.enabled = false
    end
    construct = false --Don't call myself recursively (endless loop)
      configure_envelope_loop()
    construct = true

   
    change_from_tool = false
--  end

end



---------------------------------------------------------------------------------------


function fetch_pitch_envelope(manual)
--Get pitch values from the envelope and populate the tool table

  init_tables(ENV_NOTE_COLUMN)
  local song = renoise.song()
  local tool_instrument = song.selected_instrument_index
  local instrument = song.instruments[tool_instrument].sample_envelopes
  local points = #instrument.pitch.points
  local factor = 100
  note_scheme_size = instrument.pitch.length-1

--x/y values
  
  if env_multiplier == ENV_X10 then
    factor = 10
  elseif env_multiplier == ENV_X1 then
    factor = 1
  end
  
  if env_note_value[0] == NOTE_SCHEME_TERMINATION then
    env_note_value[0] = EMPTY_CELL
  end
  
  for t = 1, points do
    local value = tonumber(instrument.pitch.points[t].value)
    local location = tonumber(instrument.pitch.points[t].time) - 1
    if location > (instrument.pitch.length-1) then
      break
    end
    if value > 0.5 then
      value = value - 0.5
      env_note_value[location] = math.round(value/TONE)
    elseif value < 0.5 then
      env_note_value[location] = math.round((0.5 - value)/TONE) *-1
    else
      env_note_value[location] = 0
    end
    env_note_value[location] = env_note_value[location] / factor
    env_note_value[location] = math.round(env_note_value[location])

  end
  
  env_note_value[instrument.pitch.length-1] = NOTE_SCHEME_TERMINATION
--envelope-type
  env_pitch_type = instrument.pitch.play_mode
--loop-mode and points
  note_loop_type = instrument.pitch.loop_mode
  
  if  note_loop_type > ENV_LOOP_OFF then
    note_loop_start = instrument.pitch.loop_start-1
    note_loop_end = instrument.pitch.loop_end-1
  end
  
--sustain
  if instrument.pitch.sustain_enabled then
    note_sustain = instrument.pitch.sustain_position - 1
  else
    note_sustain = -1
  end



--LFO data
  note_lfo1_type = instrument.pitch.lfo1.mode
  note_lfo1_phase = instrument.pitch.lfo1.phase
  note_lfo1_freq = instrument.pitch.lfo1.frequency
  note_lfo1_amount = instrument.pitch.lfo1.amount
  note_lfo2_type = instrument.pitch.lfo2.mode
  note_lfo2_phase = instrument.pitch.lfo2.phase
  note_lfo2_freq = instrument.pitch.lfo2.frequency
  note_lfo2_amount = instrument.pitch.lfo2.amount

  
end


---------------------------------------------------------------------------------------


function fetch_volume_envelope(manual)
--Get volume values from the envelope and populate the tool table
  local song = renoise.song()
  local tool_instrument =song.selected_instrument_index
  local instrument = song.instruments[tool_instrument].sample_envelopes
  local points = #instrument.volume.points
  local factor = 100
  vol_scheme_size = instrument.volume.length-1
  init_tables(ENV_VOL_COLUMN)

  envelope_volume_toggle = instrument.volume.enabled

--envelope-type
  env_volume_type = instrument.volume.play_mode
--x/y values
  for t = 0, vol_scheme_size do
    if env_vol_value[t] == VOL_PAN_TERMINATION then
      env_vol_value[t] = EMPTY_CELL
    end
  end

  for t = 1, points do
    local value = tonumber(instrument.volume.points[t].value)
    local location = tonumber(instrument.volume.points[t].time) - 1
    if location > (instrument.volume.length-1) then
      break
    end

    env_vol_value[location] = math.round(value *100)

  end
  
  --Set end of frame marker if it does not contain a value
  if env_vol_value[instrument.volume.length-1] == EMPTY_CELL then
    env_vol_value[instrument.volume.length-1] = VOL_PAN_TERMINATION
  end


--release
  vol_release = instrument.volume.fade_amount
  
--loop-mode and points
  vol_loop_type = instrument.volume.loop_mode
  if vol_loop_type > ENV_LOOP_OFF then
    vol_loop_start = instrument.volume.loop_start -1
    vol_loop_end = instrument.volume.loop_end -1
  end
--sustain
  if instrument.volume.sustain_enabled then
    vol_sustain = instrument.volume.sustain_position -1
  else
    vol_sustain = -1
  end

--lfo
  vol_lfo1_type = instrument.volume.lfo1.mode
  vol_lfo1_phase = instrument.volume.lfo1.phase
  vol_lfo1_freq = instrument.volume.lfo1.frequency
  vol_lfo1_amount = instrument.volume.lfo1.amount
  vol_lfo2_type = instrument.volume.lfo2.mode
  vol_lfo2_phase = instrument.volume.lfo2.phase
  vol_lfo2_freq = instrument.volume.lfo2.frequency
  vol_lfo2_amount = instrument.volume.lfo2.amount


  
end

---------------------------------------------------------------------------------------


function fetch_panning_envelope(manual)
--Get panning values from the envelope and populate the tool table
  
  local song = renoise.song()
  local tool_instrument = song.selected_instrument_index
  local instrument = song.instruments[tool_instrument].sample_envelopes
  local points = #instrument.pan.points
  local factor = 100
  pan_scheme_size = instrument.pan.length-1
  init_tables(ENV_PAN_COLUMN)

  envelope_panning_toggle = instrument.pan.enabled

--envelope-type
  env_panning_type = instrument.pan.play_mode
  
--x/y values
  for t = 0, pan_scheme_size do
    if env_pan_value[t] == VOL_PAN_TERMINATION then
      env_pan_value[t] = EMPTY_CELL
    end
  end

  for t = 1, points do
    local value = tonumber(instrument.pan.points[t].value)
    local location = tonumber(instrument.pan.points[t].time) - 1
    if location > (instrument.pan.length-1) then
      break
    end
    if value < 0.5 then
      value = math.round((0.5 - value) * -100)
    else
      value = math.round((value - 0.5) * 100)
    end
    env_pan_value[location] = value

  end
  if env_pan_value[instrument.pan.length-1] == EMPTY_CELL then
    env_pan_value[instrument.pan.length-1] = VOL_PAN_TERMINATION
  end
  
--loop-mode and points
  pan_loop_type = instrument.pan.loop_mode

  if pan_loop_type > ENV_LOOP_OFF then
    pan_loop_start = instrument.pan.loop_start -1
    pan_loop_end = instrument.pan.loop_end -1
  end
--sustain
  if instrument.pan.sustain_enabled then
    pan_sustain = instrument.pan.sustain_position - 1
  else
    pan_sustain = -1
  end

--lfo
  pan_lfo1_type = instrument.pan.lfo1.mode
  pan_lfo1_phase = instrument.pan.lfo1.phase
  pan_lfo1_freq = instrument.pan.lfo1.frequency
  pan_lfo1_amount = instrument.pan.lfo1.amount
  pan_lfo2_type = instrument.pan.lfo2.mode
  pan_lfo2_phase = instrument.pan.lfo2.phase
  pan_lfo2_freq = instrument.pan.lfo2.frequency
  pan_lfo2_amount = instrument.pan.lfo2.amount
    
  
end


---------------------------------------------------------------------------------------

function fetch_cut_res_envelopes()
--Get cuttoff and resonance values from the envelope and lfo
  local song = renoise.song()
  local tool_instrument = song.selected_instrument_index
  local instrument = song.instruments[tool_instrument].sample_envelopes
  local cut_points = #instrument.cutoff.points
  local res_points = #instrument.resonance.points

  cut_scheme_size = instrument.cutoff.length
  res_scheme_size = instrument.resonance.length
  env_cut_scheme = ''
  cutoff_point_scheme = ''
  env_res_scheme = ''
  resonance_point_scheme = ''
  init_tables(false)

--Filter type
  cutres_filter_type = instrument.filter_type

--envelope-type
  cutoff_enabled = instrument.cutoff.enabled
  env_cutoff_type = instrument.cutoff.play_mode

  resonance_enabled = instrument.resonance.enabled
  env_resonance_type = instrument.resonance.play_mode

--x/y values

  for t = 1, cut_points do
    local value = tonumber(instrument.cutoff.points[t].value)
    local location = tonumber(instrument.cutoff.points[t].time) - 1
    if location > (instrument.cutoff.length-1) then
      break
    end
    
    env_cut_value[location] = value
    if t == 1 then
      env_cut_scheme = tostring(value)
      cutoff_point_scheme = tostring(location)
    
    else
      env_cut_scheme = env_cut_scheme..','..tostring(value)
      cutoff_point_scheme = cutoff_point_scheme..','..tostring(location)
    
    end
  end
  
  for t = 1, res_points do
    local value = tonumber(instrument.resonance.points[t].value)
    local location = tonumber(instrument.resonance.points[t].time) - 1
    if location > (instrument.resonance.length-1) then
      break
    end
    
    env_res_value[location] = value
    if t == 1 then
      env_res_scheme = tostring(value)
      resonance_point_scheme = tostring(location)
    
    else
      env_res_scheme = env_res_scheme..','..tostring(value)
      resonance_point_scheme = resonance_point_scheme..','..tostring(location)
    
    end
  end

  
--loop-mode and points
  cut_loop_type = instrument.cutoff.loop_mode
  cut_loop_start = instrument.cutoff.loop_start -1
  cut_loop_end = instrument.cutoff.loop_end -1

  res_loop_type = instrument.resonance.loop_mode
  res_loop_start = instrument.resonance.loop_start -1
  res_loop_end = instrument.resonance.loop_end -1

--sustain
  if instrument.cutoff.sustain_enabled then
    cut_sustain = instrument.cutoff.sustain_position
  else
    cut_sustain = -1
  end

  if instrument.resonance.sustain_enabled then
    res_sustain = instrument.resonance.sustain_position
  else
    res_sustain = -1
  end

--lfo
  cut_lfo_type = instrument.cutoff.lfo.mode
  cut_lfo_phase = instrument.cutoff.lfo.phase
  cut_lfo_freq = instrument.cutoff.lfo.frequency
  cut_lfo_amount = instrument.cutoff.lfo.amount
  cut_follow = instrument.cutoff.follower.enabled
  cut_follow_attack = instrument.cutoff.follower.attack
  cut_follow_release = instrument.cutoff.follower.release
  cut_follow_amount = instrument.cutoff.follower.amount

  res_lfo_type = instrument.resonance.lfo.mode
  res_lfo_phase = instrument.resonance.lfo.phase
  res_lfo_freq = instrument.resonance.lfo.frequency
  res_lfo_amount = instrument.resonance.lfo.amount
  res_follow = instrument.resonance.follower.enabled
  res_follow_attack = instrument.resonance.follower.attack
  res_follow_release = instrument.resonance.follower.release
  res_follow_amount = instrument.resonance.follower.amount
  
  --when saving, these values should be read from the envelopes!
  note_lfo1_type = instrument.pitch.lfo1.mode
  note_lfo1_phase = instrument.pitch.lfo1.phase
  note_lfo1_freq = instrument.pitch.lfo1.frequency
  note_lfo1_amount = instrument.pitch.lfo1.amount
  note_lfo2_type = instrument.pitch.lfo2.mode
  note_lfo2_phase = instrument.pitch.lfo2.phase
  note_lfo2_freq = instrument.pitch.lfo2.frequency
  note_lfo2_amount = instrument.pitch.lfo2.amount
    
  vol_lfo1_type = instrument.volume.lfo1.mode
  vol_lfo1_phase = instrument.volume.lfo1.phase
  vol_lfo1_freq = instrument.volume.lfo1.frequency
  vol_lfo1_amount = instrument.volume.lfo1.amount
  vol_lfo2_type = instrument.volume.lfo2.mode
  vol_lfo2_phase = instrument.volume.lfo2.phase
  vol_lfo2_freq = instrument.volume.lfo2.frequency
  vol_lfo2_amount = instrument.volume.lfo2.amount  
  
  pan_lfo1_type = instrument.pan.lfo1.mode
  pan_lfo1_phase = instrument.pan.lfo1.phase
  pan_lfo1_freq = instrument.pan.lfo1.frequency
  pan_lfo1_amount = instrument.pan.lfo1.amount
  pan_lfo2_type = instrument.pan.lfo2.mode
  pan_lfo2_phase = instrument.pan.lfo2.phase
  pan_lfo2_freq = instrument.pan.lfo2.frequency
  pan_lfo2_amount = instrument.pan.lfo2.amount
    

end

---------------------------------------------------------------------------------------


function get_current_data(column)
--Prepare concatenated strings for saving by the preset manager.
  local song = renoise.song()
  local tool_instrument = song.selected_instrument_index
  local instrument = song.instruments[tool_instrument].sample_envelopes
  
  env_pitch_scheme = ''
  note_point_scheme = ''
  env_vol_scheme = ''
  volume_point_scheme = ''
  env_pan_scheme = ''
  panning_point_scheme = ''

  fetch_cut_res_envelopes() 
  envelope_pitch_toggle =  instrument.pitch.enabled
  envelope_volume_toggle = instrument.volume.enabled
  envelope_panning_toggle = instrument.pan.enabled
  
  for t = 0, note_scheme_size do
    local table_number = 0
    if string.sub(env_note_value[t],1,1) ~= '[' then
      table_number = tonumber(env_note_value[t])
    end
    if table_number < EMPTY_CELL then
      if t == 0 then
        env_pitch_scheme = tostring(env_note_value[t])
        note_point_scheme = tostring(t)
      else
        env_pitch_scheme = env_pitch_scheme .. ',' .. tostring(env_note_value[t])
        note_point_scheme = note_point_scheme .. ',' .. tostring(t)
      end
      if table_number == NOTE_SCHEME_TERMINATION then
        break
      end
    end
  end
    
  if vol_scheme_size ~= nil then
    for t = 0, vol_scheme_size do
      if tonumber(env_vol_value[t]) < EMPTY_CELL then
        if t == 0 then
          env_vol_scheme = tostring(env_vol_value[t])
          volume_point_scheme = tostring(t)
        else
          env_vol_scheme = env_vol_scheme .. ',' .. tostring(env_vol_value[t])
          volume_point_scheme = volume_point_scheme .. ',' .. tostring(t)
        end
        if tonumber(env_vol_value[t]) == VOL_PAN_TERMINATION then
          break
        end
        
      end
    end
  end
    
  if pan_scheme_size ~= nil then
    for t = 0, pan_scheme_size do
      if tonumber(env_pan_value[t]) < EMPTY_CELL then
        if t == 0 then
          env_pan_scheme = tostring(env_pan_value[t])
          panning_point_scheme = tostring(t)
        else
          env_pan_scheme = env_pan_scheme .. ',' .. tostring(env_pan_value[t])
          panning_point_scheme = panning_point_scheme .. ',' .. tostring(t)
        end
        if tonumber(env_pan_value[t]) == VOL_PAN_TERMINATION then
          break
        end
      end
    end
  end

  note_loop_type = instrument.pitch.loop_mode
  env_pitch_type = instrument.pitch.play_mode

  if instrument.pitch.sustain_enabled then
    note_sustain = instrument.pitch.sustain_position - 1
  else
    note_sustain = -1
  end

  if column == nil then
    vol_loop_type = instrument.volume.loop_mode
    env_volume_type = instrument.volume.play_mode
    vol_release = instrument.volume.fade_amount    
    vol_loop_start = instrument.volume.loop_start - 1
    vol_loop_end = instrument.volume.loop_end - 1

    if instrument.volume.sustain_enabled then
      vol_sustain = instrument.volume.sustain_position - 1
    else
      vol_sustain = -1
    end

    pan_loop_type = instrument.pan.loop_mode
    env_panning_type = instrument.pan.play_mode
    pan_loop_start = instrument.pan.loop_start - 1
    pan_loop_end = instrument.pan.loop_end - 1
    
    if instrument.pan.sustain_enabled then
      pan_sustain = instrument.pan.sustain_position - 1
    else
      pan_sustain = -1
    end  
    
  end
    
end



---------------------------------------------------------------------------------------

function apply_unattended_properties()
  local song = renoise.song()
  local tool_instrument = song.selected_instrument_index
  local instrument = song.instruments[tool_instrument].sample_envelopes

  disable_cut_res_lfo()
  
  set_note_lfo_data(instrument)
  set_vol_lfo_data(instrument)
  set_pan_lfo_data(instrument)
  set_cut_lfo_data(instrument)
  set_res_lfo_data(instrument)
  set_cutoff_data(instrument)
  set_resonance_data(instrument)
  
end

---------------------------------------------------------------------------------------

function disable_cut_res_lfo()

  local song = renoise.song()
  local tool_instrument = song.selected_instrument_index
  local instrument = song.instruments[tool_instrument].sample_envelopes
  instrument.pitch.lfo1.mode = 1
  instrument.pitch.lfo2.mode = 1
  instrument.volume.lfo1.mode = 1
  instrument.volume.lfo2.mode = 1
  instrument.pan.lfo1.mode = 1
  instrument.pan.lfo2.mode = 1

  instrument.cutoff.enabled = false
  instrument.cutoff.lfo.mode = 1
  instrument.cutoff.follower.enabled = false
  instrument.resonance.enabled = false
  instrument.resonance.lfo.mode = 1 
  instrument.resonance.follower.enabled = false
  
end

---------------------------------------------------------------------------------------

function set_note_lfo_data(instrument)
  instrument.pitch.lfo1.mode = note_lfo1_type
  instrument.pitch.lfo1.phase = note_lfo1_phase
  instrument.pitch.lfo1.frequency = note_lfo1_freq
  instrument.pitch.lfo1.amount = note_lfo1_amount
  instrument.pitch.lfo2.mode = note_lfo2_type
  instrument.pitch.lfo2.phase = note_lfo2_phase
  instrument.pitch.lfo2.frequency = note_lfo2_freq
  instrument.pitch.lfo2.amount = note_lfo2_amount
end

function set_vol_lfo_data(instrument)
  instrument.volume.lfo1.mode = vol_lfo1_type
  instrument.volume.lfo1.phase = vol_lfo1_phase
  instrument.volume.lfo1.frequency = vol_lfo1_freq
  instrument.volume.lfo1.amount = vol_lfo1_amount
  instrument.volume.lfo2.mode = vol_lfo2_type
  instrument.volume.lfo2.phase = vol_lfo2_phase
  instrument.volume.lfo2.frequency = vol_lfo2_freq
  instrument.volume.lfo2.amount = vol_lfo2_amount
end

function set_pan_lfo_data(instrument)
  instrument.pan.lfo1.mode = pan_lfo1_type
  instrument.pan.lfo1.phase = pan_lfo1_phase
  instrument.pan.lfo1.frequency = pan_lfo1_freq
  instrument.pan.lfo1.amount = pan_lfo1_amount
  instrument.pan.lfo2.mode = pan_lfo2_type
  instrument.pan.lfo2.phase = pan_lfo2_phase
  instrument.pan.lfo2.frequency = pan_lfo2_freq
  instrument.pan.lfo2.amount = pan_lfo2_amount
end

function set_cut_lfo_data(instrument)
  if instrument.filter_type ~= cutres_filter_type then
    instrument.filter_type = cutres_filter_type
  end
  instrument.cutoff.lfo.mode = cut_lfo_type 
  instrument.cutoff.lfo.phase = cut_lfo_phase 
  instrument.cutoff.lfo.frequency = cut_lfo_freq
  instrument.cutoff.lfo.amount = cut_lfo_amount
  instrument.cutoff.follower.enabled = toboolean(cut_follow)
-- First change the release, then the attack, due to a bug in the LFO's value changing routine
  if cut_follow_release > 0 then
    instrument.cutoff.follower.release = cut_follow_release
    instrument.cutoff.follower.attack = cut_follow_attack
    instrument.cutoff.follower.amount = cut_follow_amount
  end
end

function set_res_lfo_data(instrument)
  if instrument.filter_type ~= cutres_filter_type then
    instrument.filter_type = cutres_filter_type
  end
  instrument.resonance.lfo.mode = res_lfo_type 
  instrument.resonance.lfo.phase = res_lfo_phase 
  instrument.resonance.lfo.frequency = res_lfo_freq
  instrument.resonance.lfo.amount = res_lfo_amount
  instrument.resonance.follower.enabled = toboolean(res_follow)
-- First change the release, then the attack, due to a bug in the LFO's value changing routine
  if res_follow_release > 0 then
    instrument.resonance.follower.release = res_follow_release
    instrument.resonance.follower.attack = res_follow_attack
    instrument.resonance.follower.amount = res_follow_amount
  end
end

---------------------------------------------------------------------------------------

function set_cutoff_data(instrument)
  if instrument.filter_type ~= cutres_filter_type then
    instrument.filter_type = cutres_filter_type
  end
  
  if cut_scheme_size >= 6 then
    instrument.cutoff.length = cut_scheme_size
  else
    instrument.cutoff.length = 6
  end  
  
  instrument.cutoff.play_mode = env_cutoff_type
  instrument.cutoff.enabled = toboolean(cutoff_enabled)
  instrument.cutoff.loop_mode = cut_loop_type

  if cut_loop_type > ENV_LOOP_OFF then
    instrument.cutoff.loop_start = cut_loop_start + 1
    instrument.cutoff.loop_end = cut_loop_end + 1
  end
  if cut_sustain > -1 then
    instrument.cutoff.sustain_enabled = true
    instrument.cutoff.sustain_position = cut_sustain
  else
    instrument.cutoff.sustain_enabled = false
  end

  instrument.cutoff:clear_points()

  for x = 0,cut_scheme_size do
    if env_cut_value[x] ~= nil then
      if tonumber(env_cut_value[x]) < VOL_PAN_TERMINATION then
        local cut_value = tonumber(env_cut_value[x])
        instrument.cutoff:add_point_at(x+1,cut_value)
      end
    end
  end  

end

---------------------------------------------------------------------------------------

function set_resonance_data(instrument)
  if instrument.filter_type ~= cutres_filter_type then
    instrument.filter_type = cutres_filter_type
  end

  if res_scheme_size >= 6 then
    instrument.resonance.length = res_scheme_size
  else
    instrument.resonance.length = 6
  end

  instrument.resonance.play_mode = env_resonance_type
  instrument.resonance.enabled = toboolean(resonance_enabled)
  instrument.resonance.loop_mode = res_loop_type

  if res_loop_type > ENV_LOOP_OFF then
    instrument.resonance.loop_start = res_loop_start + 1
    instrument.resonance.loop_end = res_loop_end + 1
  end
  if res_sustain > -1 then
    instrument.resonance.sustain_enabled = true
    instrument.resonance.sustain_position = res_sustain
  else
    instrument.resonance.sustain_enabled = false
  end

  instrument.resonance:clear_points()

  for x = 0,res_scheme_size do
    if env_res_value[x] ~= nil then
      if tonumber(env_res_value[x]) < VOL_PAN_TERMINATION then
        local res_value = tonumber(env_res_value[x])
        instrument.resonance:add_point_at(x+1,res_value)
      end
    end
  end  

end

---------------------------------------------------------------------------------------

function populate_cutres_tables(envelope)
    local y_value_table = {}
    local x_value_table = {}
    local end_node = #y_value_table
    local leading_table = 1 --Notes
    local sub_position = 1
        
    if envelope == ENV_CUT then
      y_value_table = env_cut_scheme:split( "[^,%s]+" )
      x_value_table = cutoff_point_scheme:split( "[^,%s]+" )

    elseif envelope == ENV_RES then
      y_value_table = env_res_scheme:split( "[^,%s]+" )
      x_value_table = resonance_point_scheme:split( "[^,%s]+" )

    end
          
    end_node = #y_value_table
          
    if #y_value_table < #x_value_table then
      end_node = #x_value_table
      leading_table = 2
    end
          
    if #y_value_table == #x_value_table then
      leading_table = -1
    end
          
    if leading_table == 1 then
      sub_position = 2
            
      for t = #x_value_table+1, end_node do
        local new_val = tonumber(x_value_table[sub_position]) -
                        tonumber(x_value_table[sub_position -1])
        x_value_table[t] = tonumber(x_value_table[t-1])+new_val
        sub_position = sub_position + 1
      end
    else
      for t = #y_value_table+1, end_node do
        y_value_table[t] = y_value_table[sub_position]
        sub_position = sub_position + 1
        if sub_position > #y_value_table then
          sub_position = 1
        end
      end
    end
    
    for t = 1, end_node do
      if y_value_table[t] ~= nil and x_value_table[t] ~= nil then
        if envelope == ENV_CUT then
          env_cut_value[tonumber(x_value_table[t])] = y_value_table[t]
        elseif envelope == ENV_RES then
          env_res_value[tonumber(x_value_table[t])] = y_value_table[t]
        end
      end
    end

end



--------------------------------------------------------------------------------
--      End of teh road...
--------------------------------------------------------------------------------

  
