--[[============================================================================
globals.lua
============================================================================]]--

--require "remdebug.engine"
MENU_TITLE = "Envelope Preset Manager..."
TOOL_TITLE = "Instrument envelope preset manager"
TOOL_FILE = ""
--Renoise internal boundary constants
NUM_OCTAVES = 10         --:0 to 9   = 10
NUM_NOTES = 12

MAX_TICKS = 15           --:0 to 15  = 16
MAX_DELAY_STEPS = 255    --:0 to 255 = 256
MAX_NOTE_COLUMNS = 12
MAX_EFFECT_COLUMNS = 8
EMPTY = 255              -- value to clear Instrument, Panning, volume and delay column
EMPTY_NOTE = 121

--Initialization
arpeg_option_dialog = nil
tone_mode = 1
first_show = false

--Gui object definitions
DIALOG_MARGIN = renoise.ViewBuilder.DEFAULT_DIALOG_MARGIN
CONTENT_MARGIN = renoise.ViewBuilder.DEFAULT_CONTROL_MARGIN
CONTENT_SPACING = renoise.ViewBuilder.DEFAULT_CONTROL_SPACING
SECTION_MARGIN = 1
TOOL_FRAME_WIDTH = 290
CHECKBOX_WIDTH = 30
TEXT_ROW_WIDTH = 80
ea_gui = renoise.ViewBuilder()
tool_dialog = nil
change_from_tool = false

--------------------------------------------------
-- Envelope arpeggiator options
--------------------------------------------------
preset_area_height = 52
preset_options_height = 190
preset_version = "3.20"

processing_instrument = 1
note_scheme_size  = -1
vol_scheme_size  = -1
pan_scheme_size  = -1
cut_scheme_size = -1
res_scheme_size = -1

EMPTY_CELL = 9999
NOTE_OFF = 2555
NOTE_SCHEME_TERMINATION = 2222
VOL_PAN_TERMINATION = 256
MINIMUM_FRAME_LENGTH = 5
MAXIMUM_FRAME_LENGTH = 1000

CURSOR_POS_COLOR = {0x80,0x80,0x80}
CURSOR_POS_EDIT_COLOR = {0xff,0x00,0x00}
COLOR_UNSELECTED = {0x01,0x00,0x00}
COLOR_THEME = {0x00,0x00,0x00}
LOOP_START_COLOR = {0x00,0x84,0x0c}
LOOP_END_COLOR = {0x00,0x1d,0xff}
SUSTAIN_COLOR = {0xff,0xff,0xff}

IMAGE_UNTOGGLED_MODE = 'body_color'
IMAGE_TOGGLED_MODE = 'button_color'

ENV_NOTE_COLUMN = 1
ENV_VOL_COLUMN = 2
ENV_PAN_COLUMN = 3 
ENV_CUT = 4
ENV_RES = 5
ROW_HEIGHT = 12

ARP_MODE_OFF = 1
ARP_MODE_AUTO = 2
ARP_MODE_MANUAL = 3

FREQ_TYPE_LINES = 1
FREQ_TYPE_POINTS = 2
FREQ_TYPE_FREEFORM = 3

ENV_TYPE_POINTS = 1
ENV_TYPE_LINEAIR = 2
ENV_TYPE_CURVE = 3

ENV_LOOP_OFF = 1
ENV_LOOP_FORWARD = 2
ENV_LOOP_BACKWARD = 3
ENV_LOOP_PINGPONG = 4

LOOP_START = 1
LOOP_END = 2
ENV_LOOP_TYPE = {'off','>>', '<<', '><'}

env_current_line ={}
env_current_line.col = ENV_NOTE_COLUMN
env_current_line.row = 0

loop_pos_color = {}
note_pos_color = {}
vol_pos_color = {}
pan_pos_color = {}
line_position_offset = 0
visible_lines = 15
env_pattern_line = {}
note_pos_color[env_current_line.row]= CURSOR_POS_COLOR
envelope_pitch_toggle = false
envelope_volume_toggle = false
envelope_panning_toggle = false


env_note_value = {}
env_pitch_scheme = ''
note_point_scheme = ''
note_freq_val = 0
note_freq_type = FREQ_TYPE_FREEFORM
env_pitch_type = ENV_TYPE_POINTS --should be fetched from envelope!
auto_note_loop = ARP_MODE_OFF
note_loop_type = 1
note_loop_start = -1
note_loop_end = -1
note_sustain = -1
note_lfo_data = false
note_lfo1_type = 1
note_lfo1_phase = 0
note_lfo1_freq = 1
note_lfo1_amount = 16
note_lfo2_type = 1
note_lfo2_phase = 0
note_lfo2_freq = 1
note_lfo2_amount = 16


env_vol_value = {}
env_vol_scheme = ''
volume_point_scheme = ''
vol_freq_val = 0
vol_freq_type = FREQ_TYPE_FREEFORM
env_volume_type = ENV_TYPE_POINTS --should be fetched from envelope!
auto_vol_loop = ARP_MODE_OFF --Autoloop support yes/no?
vol_loop_type = 1 --The type of loop from the drop down
vol_loop_start = -1
vol_loop_end = -1
vol_sustain = -1
vol_release = -1
vol_assist_low_val = 0
vol_assist_high_val = 100
vol_assist_high_size = 0
vol_pulse_mode = ARP_MODE_OFF
vol_lfo_data = false
vol_lfo1_type = 1
vol_lfo1_phase = 0
vol_lfo1_freq = 1
vol_lfo1_amount = 16
vol_lfo2_type = 1
vol_lfo2_phase = 0
vol_lfo2_freq = 1
vol_lfo2_amount = 16


env_pan_value = {}
env_pan_scheme = ''
panning_point_scheme = ''
pan_freq_val = 0
pan_freq_type = FREQ_TYPE_FREEFORM
env_panning_type = ENV_TYPE_POINTS --should be fetched from envelope!
auto_pan_loop = ARP_MODE_OFF --Autoloop support yes/no?
pan_loop_type = 1 --The type of loop from the drop down
pan_loop_start = -1
pan_loop_end = -1
pan_sustain = -1
pan_assist_first_val = 0
pan_assist_next_val = 0
pan_assist_first_size = 0
pan_pulse_mode = ARP_MODE_OFF
pan_lfo_data = false
pan_lfo1_type = 1
pan_lfo1_phase = 0
pan_lfo1_freq = 1
pan_lfo1_amount = 16
pan_lfo2_type = 1
pan_lfo2_phase = 0
pan_lfo2_freq = 1
pan_lfo2_amount = 16

cutoff_data = false
cutoff_enabled = false
env_cut_value = {}
env_cut_scheme = ''
cutoff_point_scheme = ''
env_cutoff_type = ENV_TYPE_CURVE
cut_loop_type = 1
cut_loop_start = -1
cut_loop_end = -1
cut_sustain  = -1
cut_lfo_data = false
cut_lfo_type = 1
cut_lfo_phase = 0
cut_lfo_freq = 1
cut_lfo_amount = 16
cut_follow = 2
cut_follow_attack = 0
cut_follow_release = 0
cut_follow_amount = 0


resonance_data = false
resonance_enabled = false
env_res_value = {}
env_res_scheme = ''
resonance_point_scheme = ''
resonance_filter_type = 1
env_resonance_type = ENV_TYPE_CURVE
res_loop_type = 1
res_loop_start = -1
res_loop_end = -1
res_sustain  = -1
res_lfo_data = false
res_lfo_type = 1
res_lfo_phase = 0
res_lfo_freq = 1
res_lfo_amount = 16
res_follow = 2
res_follow_attack = 0
res_follow_release = 0
res_follow_amount = 0

cutres_filter_type = 1

note_mode = true --Show notes or figures in the column?


PLAY_MODE_POINTS = 1
ENV_X1 = 1
ENV_X10 = 2
ENV_X100 = 3
TONE = 0.5/1200
pitch_table = {}

tone_factor = ENV_X1
transpose_pitch_scheme = false

env_multiplier = ENV_X1

env_sync_mode = true
--env_free_sync_mode = false

env_auto_apply = true
construct = true  --Execute the construct function by default

change_from_renoise = nil


--------------------------------------------------
-- Generic options


--Preset manager
PATH_SEPARATOR = "/"
TOOL_SAVE_FOLDER = "__tools_save_data__"
print(renoise.tool().bundle_path)
if (os.platform() == "WINDOWS") then
  PATH_SEPARATOR = "\\"
end

env_preset_field = nil
env_preset_file = nil
env_preset_name = ""
env_preset_list = {}
env_preset_table = {"PitchApply","ToneFactor",
  "NoteDrawType","NoteScheme","NotePointSequences","NoteSchemeSize",
  "NoteLoopType","NoteLoopStart","NoteLoopEnd","NoteSustain",
  "NoteLFO1","NoteLFOPhase1","NoteLFOFrequency1", "NoteLFOAmount1",
  "NoteLFO2","NoteLFOPhase2","NoteLFOFrequency2", "NoteLFOAmount2",

  "VolumeScheme","VolumePointSequences","VolumeSchemeSize","VolumeDrawType",
  "VolumeFrequency","VolumeFrequencyMode","VolumeApply","VolumeLoopType",
  "VolumeLoopStart","VolumeLoopEnd","VolumeSustain","VolumeRelease",
  "VolumeLFO1","VolumeLFOPhase1","VolumeLFOFrequency1", "VolumeLFOAmount1",
  "VolumeLFO2","VolumeLFOPhase2","VolumeLFOFrequency2", "VolumeLFOAmount2",
  
  "PanningScheme","PanningPointSequences","PanningSchemeSize","PanningDrawType",
  "PanningFrequency","PanningFrequencyMode","PanningApply",
  "PanningLoopType","PanningLoopStart","PanningLoopEnd","PanningSustain",
  "PanningLFO1","PanningLFOPhase1","PanningLFOFrequency1", "PanningLFOAmount1",
  "PanningLFO2","PanningLFOPhase2","PanningLFOFrequency2", "PanningLFOAmount2",

  "CutoffScheme","CutoffPointSequences","CutoffSchemeSize",
  "CutoffDrawType","CutoffEnabled",
  "CutoffLoopType","CutoffLoopStart","CutoffLoopEnd","CutoffSustain",
  "CutoffLFO","CutoffLFOPhase","CutoffLFOFrequency", "CutoffLFOAmount",
  "CutoffFollower","CutoffFollowerAttack","CutoffFollowerRelease", "CutoffFollowerAmount",

  "ResonanceScheme","ResonancePointSequences","ResonanceSchemeSize",
  "ResonanceDrawType","ResonanceEnabled",
  "ResonanceLoopType","ResonanceLoopStart","ResonanceLoopEnd","ResonanceSustain",
  "ResonanceLFO","ResonanceLFOPhase","ResonanceLFOFrequency", "ResonanceLFOAmount",
  "ResonanceFollower","ResonanceFollowerAttack","ResonanceFollowerRelease", "ResonanceFollowerAmount",

  "CutResFilterType",
  }

--------------------------------------------------------------------------------
-- helper functions
--------------------------------------------------------------------------------

function randomize(tstart, tend, buffer, compare)
  local number = tostring(os.clock())
  local approved = false
  if tstart == 0 then
    tstart = 1
  end
  if tend < tstart then
    tend = tstart
  end
  if string.find(number,"%.") ~= nil then
    number = string.sub(number, string.find(number,"%.")+1)
  end
  
  math.randomseed( tonumber(number))
   number  = number + math.random(1, 30)
   math.randomseed( tonumber(number))
   math.random(tstart, tend); math.random(tstart, tend); math.random(tstart, tend)
  local result = math.random(tstart, tend)
  return result
end

--------------------------------------------------------------------------------

function math.round(value)
  local floor_val = math.floor(value)
  local float = value - floor_val
  if float < 0.5 then 
    return math.floor(value)
  else
    return math.ceil(value)
  end
end

--------------------------------------------------------------------------------

string.split = function(str, pattern)
  pattern = pattern or "[^%s]+"

  if pattern:len() == 0 then pattern = "[^%s]+" end

  local parts = {__index = table.insert}
  setmetatable(parts, parts)
  str:gsub(pattern, parts)
  setmetatable(parts, nil)
  parts.__index = nil
  return parts
end

--------------------------------------------------------------------------------

function get_tools_root()    
  local dir = renoise.tool().bundle_path
  return dir:sub(1,dir:find("Tools")+5)      
end
--------------------------------------------------------------------------------

function get_tools_file()    
  local dir = renoise.tool().bundle_path
  return dir:sub(dir:find("Tools")+6,string.len(dir)-1)      
end

--------------------------------------------------------------------------------

function toboolean(value)
  if string.lower(tostring(value)) == "false" or tonumber(value) == 0 then
    return false
  end
  
  return true
end

--------------------------------------------------------------------------------

table.serialize = function(table, separator)
  local stream = nil
  for _ = 1, #table do
    if stream == nil then
      stream = table[_]
    else
      stream = stream .. separator.. table[_]
    end
  end
  return stream  
end



--------------------------------------------------------------------------------
--      End of teh road...
--------------------------------------------------------------------------------




