--[[============================================================================
gui_envelope_arp.lua, the complete layout for the envelope arpeggio components.
============================================================================]]--



function envelope_preset_dialog(vb)
  local env_pattern_header = vb:row {}
  --local env_pattern_line = vb:row {}
  local env_pattern_track = vb:column {}
  local env_pattern_control = vb:row {}
  local env_pattern_fetch = vb:row {}
  local env_pattern_sync = vb:row {}
   
  local scheme_area = vb:column{
    margin = CONTENT_MARGIN,
    spacing = CONTENT_SPACING,
    width = TOOL_FRAME_WIDTH,
    id='envelope_track',

    vb:horizontal_aligner {
      mode = "left",
      width = "100%",
      margin=SECTION_MARGIN,
      
      vb:column{
        id='right_column',
        vb:vertical_aligner {
          id='envelope_preset_area',
--          mode='left',
          vb:space{height=2},
          vb:row{
            vb:space{width=2},
            envelope_preset_management(vb),
          },
        },
      }
    }
  }
  return scheme_area
    
end


---------------------------------------------------------------------------------------

function envelope_preset_management(vb)
  -------------------------------------------------------------------------
  -- Envelope Preset management
  
  local preset_area=vb:column {
    vb:row {
      vb:text {
--        width=TEXT_ROW_WIDTH,
        text='Preset'
      },
      vb:space{width=10},
      vb:popup {
        id='env_popup_preset',
        width=212,
        items=env_preset_list,
        value=1,
        tooltip="Select previously saved preset\n",
        notifier=function(value)env_preset_field = value end
      },
  
    },
    vb:horizontal_aligner {
      mode="left",
      vb:row{

        vb:button {
          width=10,
          text='Loc',
          tooltip='Show preset-folder'..
                  ' location in platform explorer',
          notifier=function()show_folder()end
        },
        vb:space {width=16},
        vb:button {
          width=10,
          text='Load',
          tooltip='Load selected preset / '..
                  'refresh presetlist (if you  have added new preset files)',
          notifier=function()
            local env_preset_file = vb.views.env_popup_preset.items[vb.views.env_popup_preset.value]
            load_preset(env_preset_file,vb)
          end
        },            
        vb:button {
          width=10,
          text='Save',
          tooltip='Save current configuration',
          notifier=function()
            get_current_data()
            save_dialog(vb)
          end
        },
      } 
    }
  }

  local property_area = vb:column{
--    margin = CONTENT_MARGIN,
  --  spacing = CONTENT_SPACING,
    --width = 310,
    --id='envelope_preset_area',
    vb:horizontal_aligner {
      mode = "justify",
      width = "100%",
      margin=0,
      vb:column {
        vb:row{
          vb:column{
            margin = DIALOG_MARGIN,
            style = "group",

            vb:horizontal_aligner {
              mode = "center",
              preset_area
            }
          }
        }
      }
    }
  }  
  return property_area
end



--------------------------------------------------------------------------------
-- Saving pattern arpeggio presets
--------------------------------------------------------------------------------

function save_dialog(vrb)
  local vb = renoise.ViewBuilder() 
  local CONTENT_MARGIN = renoise.ViewBuilder.DEFAULT_CONTROL_MARGIN
  local CONTENT_SPACING = renoise.ViewBuilder.DEFAULT_CONTROL_SPACING
  local dialog_title = "Save Envelope Preset"
  local preset_file = env_preset_name
  local saving_dialog = nil
  
  if vrb.views.env_popup_preset ~= nil then
    preset_file = vrb.views.env_popup_preset.items[vrb.views.env_popup_preset.value]
  end
  
  local dialog_content = vb:column { 
      margin = CONTENT_MARGIN,
      spacing = CONTENT_SPACING,
      uniform = true,
      vb:text {
        align="center",
        text = "Preset name"
      },
      vb:textfield {
        id='textfield_filename',
        width=230,
        value=preset_file,
        tooltip="enter presetname without extensions",
        notifier=function(value)
          preset_file = value 
          preset_file = string.gsub(preset_file,",", "_")
          preset_file = string.gsub(preset_file," ", "_")
        end
      },
      vb:horizontal_aligner{
        mode = "center",
        vb:row{
          spacing = 8,
          vb:button {
            width=50,
            text='Save',
            tooltip='Save the preset under this name (existing files will be overwritten'..
                    ' without warning!!',
            notifier=function()
              save_preset(preset_file,vrb)
              
              saving_dialog:close()
              saving_dialog = nil
            end
          },
          vb:button {
            width=50,
            text='Cancel',
            tooltip='Abort saving',
            notifier=function()
              saving_dialog:close()
              saving_dialog = nil
            end
          },
        }
      }
    }
  if (saving_dialog and saving_dialog.visible)then
    saving_dialog:show()
  else
    saving_dialog = nil
    saving_dialog = renoise.app():show_custom_dialog(dialog_title,
    dialog_content)
  end

end


function file_exists_dialog(preset_name)
  local vb = renoise.ViewBuilder()

  local DIALOG_MARGIN = renoise.ViewBuilder.DEFAULT_DIALOG_MARGIN
  local DIALOG_SPACING = renoise.ViewBuilder.DEFAULT_DIALOG_SPACING
  local message = "Preset "..preset_name.." already exists... Overwrite?"
  local choice = renoise.app():show_prompt("Warning",message, {"Overwrite", "Cancel"})

  if choice == 1 or choice == "Overwrite" then 
    return true
  else
    return false
  end
end


