
--[[============================================================================
preset_manager.lua
============================================================================]]--

class "PresetTable" (renoise.Document.DocumentNode)
  function PresetTable:__init()    
    renoise.Document.DocumentNode.__init(self) 
    self:add_property("PresetVersion", "2.00")
    for x = 1,#pat_preset_table do
      self:add_property(pat_preset_table[x], "00")
    end
  end
  
 
class "EnvelopePresetTable" (renoise.Document.DocumentNode)
  function EnvelopePresetTable:__init()    
    renoise.Document.DocumentNode.__init(self) 
    self:add_property("PresetVersion", "3.15")
    for x = 1,#env_preset_table do
      self:add_property(env_preset_table[x], "00")
    end
  end

local env_preset_table = EnvelopePresetTable()


---------------------------------------------------------------------------------------

function get_preset_list(vb)
  local preset_root = get_tools_root()..TOOL_SAVE_FOLDER
  local env_preset_folder = preset_root..PATH_SEPARATOR..TOOL_FILE
  check_preset_folder(preset_root)
  
  local env_preset_files = os.filenames(env_preset_folder)
  
  for _ = 1, #env_preset_files do
    env_preset_files[_] = string.gsub(env_preset_files[_],".xml","")
  end
  if ea_gui.views.env_popup_preset ~= nil then
    ea_gui.views.env_popup_preset.items = env_preset_files
  end
  create_menu(env_preset_files)
end



function create_menu(no_of_presets)
  local save_entry = "Instrument Box:Save Sample Envelope As..."
  local ssave_entry = "Sample List:Save Sample Envelope As..."

  if #no_of_presets > 0 then --needed for when no vsts available 
    local imenu_entry = ""
    local smenu_entry = ""
    for i = 1,#no_of_presets do
      imenu_entry = "Instrument Box:Load Sample Envelope...:"..no_of_presets[i]
      smenu_entry = "Sample List:Load Sample Envelope...:"..no_of_presets[i]
      
      if not renoise.tool():has_menu_entry(imenu_entry) then
        renoise.tool():add_menu_entry {
          name = imenu_entry,
          invoke = function() load_preset(no_of_presets[i]) end
        }
      end
      if not renoise.tool():has_menu_entry(smenu_entry) then
        renoise.tool():add_menu_entry {
          name = smenu_entry,
          invoke = function() load_preset(no_of_presets[i]) end
        }
      end
      
    end
  else
    
    if renoise.tool():has_menu_entry("Instrument Box:Load Sample Envelope...") then
      renoise.tool():remove_menu_entry ("Instrument Box:Load Sample Envelope...")
    end
    if renoise.tool():has_menu_entry("Sample List:Load Sample Envelope...") then
      renoise.tool():remove_menu_entry("Sample List:Load Sample Envelope...")
    end
  
  end
  
  if not renoise.tool():has_menu_entry(save_entry) then
    renoise.tool():add_menu_entry {
      name = save_entry,
      invoke = function() save_dialog(ea_gui) end
    }
  end
  if not renoise.tool():has_menu_entry(ssave_entry) then
    renoise.tool():add_menu_entry {
      name = ssave_entry,
      invoke = function() save_dialog(ea_gui) end
    }
  end
end

---------------------------------------------------------------------------------------

function save_preset(preset_name,vb)
  local preset_root = get_tools_root()..TOOL_SAVE_FOLDER
  local instrument = renoise.song().instruments[renoise.song().selected_instrument_index].sample_envelopes
    env_multiplier = 1
    tone_factor = 1
    
    fetch_pitch_envelope()    
    fetch_volume_envelope()
    fetch_panning_envelope()
    fetch_cut_res_envelopes()
    get_current_data()
    env_preset_table:property("PresetVersion").value = "3.20"
    env_preset_table:property("PitchApply").value = tostring(envelope_pitch_toggle)
    env_preset_table:property("NoteScheme").value = tostring(env_pitch_scheme)
    env_preset_table:property("NotePointSequences").value = tostring(note_point_scheme)
    env_preset_table:property("NoteSchemeSize").value = tostring(note_scheme_size)
    env_preset_table:property("NoteLoopType").value = tostring(note_loop_type)
    env_preset_table:property("NoteDrawType").value = tostring(env_pitch_type)
    env_preset_table:property("NoteLoopStart").value = tostring(note_loop_start)
    env_preset_table:property("NoteLoopEnd").value = tostring(note_loop_end)
    env_preset_table:property("NoteSustain").value = tostring(note_sustain)
    env_preset_table:property("ToneFactor").value = tostring(env_multiplier)

    env_preset_table:property("NoteLFO1").value = tostring(note_lfo1_type)
    env_preset_table:property("NoteLFOPhase1").value = tostring(note_lfo1_phase)
    env_preset_table:property("NoteLFOFrequency1").value = tostring(note_lfo1_freq)
    env_preset_table:property("NoteLFOAmount1").value = tostring(note_lfo1_amount)
    env_preset_table:property("NoteLFO2").value = tostring(note_lfo2_type)
    env_preset_table:property("NoteLFOPhase2").value = tostring(note_lfo2_phase)
    env_preset_table:property("NoteLFOFrequency2").value = tostring(note_lfo2_freq)
    env_preset_table:property("NoteLFOAmount2").value = tostring(note_lfo2_amount)

    env_preset_table:property("VolumeApply").value = tostring(envelope_volume_toggle)
    env_preset_table:property("VolumeScheme").value = tostring(env_vol_scheme)
    env_preset_table:property("VolumePointSequences").value = tostring(volume_point_scheme)
    env_preset_table:property("VolumeSchemeSize").value = tostring(vol_scheme_size)
    env_preset_table:property("VolumeDrawType").value =  tostring(env_volume_type)
    env_preset_table:property("VolumeLoopType").value =  tostring(vol_loop_type)
    env_preset_table:property("VolumeLoopStart").value =  tostring(vol_loop_start)
    env_preset_table:property("VolumeLoopEnd").value =  tostring(vol_loop_end)
    env_preset_table:property("VolumeSustain").value =  tostring(vol_sustain)
    env_preset_table:property("VolumeRelease").value =  tostring(vol_release)

    env_preset_table:property("VolumeLFO1").value = tostring(vol_lfo1_type)
    env_preset_table:property("VolumeLFOPhase1").value = tostring(vol_lfo1_phase)
    env_preset_table:property("VolumeLFOFrequency1").value = tostring(vol_lfo1_freq)
    env_preset_table:property("VolumeLFOAmount1").value = tostring(vol_lfo1_amount)
    env_preset_table:property("VolumeLFO2").value = tostring(vol_lfo2_type)
    env_preset_table:property("VolumeLFOPhase2").value = tostring(vol_lfo2_phase)
    env_preset_table:property("VolumeLFOFrequency2").value = tostring(vol_lfo2_freq)
    env_preset_table:property("VolumeLFOAmount2").value = tostring(vol_lfo2_amount)

    
    env_preset_table:property("PanningApply").value = tostring(envelope_panning_toggle)
    env_preset_table:property("PanningScheme").value = tostring(env_pan_scheme)
    env_preset_table:property("PanningPointSequences").value = tostring(panning_point_scheme)
    env_preset_table:property("PanningSchemeSize").value = tostring(pan_scheme_size)
    env_preset_table:property("PanningDrawType").value =  tostring(env_panning_type)
    env_preset_table:property("PanningLoopType").value =  tostring(pan_loop_type)
    env_preset_table:property("PanningLoopStart").value =  tostring(pan_loop_start)
    env_preset_table:property("PanningLoopEnd").value =  tostring(pan_loop_end)
    env_preset_table:property("PanningSustain").value =  tostring(pan_sustain)

    env_preset_table:property("PanningFrequencyMode").value = tostring(pan_freq_type)
    env_preset_table:property("PanningFrequency").value = tostring(pan_freq_val)

    env_preset_table:property("PanningLFO1").value = tostring(pan_lfo1_type)
    env_preset_table:property("PanningLFOPhase1").value = tostring(pan_lfo1_phase)
    env_preset_table:property("PanningLFOFrequency1").value = tostring(pan_lfo1_freq)
    env_preset_table:property("PanningLFOAmount1").value = tostring(pan_lfo1_amount)
    env_preset_table:property("PanningLFO2").value = tostring(pan_lfo2_type)
    env_preset_table:property("PanningLFOPhase2").value = tostring(pan_lfo2_phase)
    env_preset_table:property("PanningLFOFrequency2").value = tostring(pan_lfo2_freq)
    env_preset_table:property("PanningLFOAmount2").value = tostring(pan_lfo2_amount)


    env_preset_table:property("CutoffScheme").value = tostring(env_cut_scheme)
    env_preset_table:property("CutoffPointSequences").value = tostring(cutoff_point_scheme)
    env_preset_table:property("CutoffSchemeSize").value = tostring(cut_scheme_size)
    env_preset_table:property("CutoffDrawType").value =  tostring(env_cutoff_type)
    env_preset_table:property("CutoffEnabled").value =  tostring(cutoff_enabled)
    env_preset_table:property("CutoffLoopType").value =  tostring(cut_loop_type)
    env_preset_table:property("CutoffLoopStart").value =  tostring(cut_loop_start)
    env_preset_table:property("CutoffLoopEnd").value =  tostring(cut_loop_end)
    env_preset_table:property("CutoffSustain").value =  tostring(cut_sustain)

    env_preset_table:property("CutoffLFO").value = tostring(cut_lfo_type)
    env_preset_table:property("CutoffLFOPhase").value = tostring(cut_lfo_phase)
    env_preset_table:property("CutoffLFOFrequency").value = tostring(cut_lfo_freq)
    env_preset_table:property("CutoffLFOAmount").value = tostring(cut_lfo_amount)
    env_preset_table:property("CutoffFollower").value = tostring(cut_follow)
    env_preset_table:property("CutoffFollowerAttack").value = tostring(cut_follow_attack)
    env_preset_table:property("CutoffFollowerRelease").value = tostring(cut_follow_release)
    env_preset_table:property("CutoffFollowerAmount").value = tostring(cut_follow_amount)


    env_preset_table:property("ResonanceScheme").value = tostring(env_res_scheme)
    env_preset_table:property("ResonancePointSequences").value = tostring(resonance_point_scheme)
    env_preset_table:property("ResonanceSchemeSize").value = tostring(res_scheme_size)
    env_preset_table:property("ResonanceDrawType").value =  tostring(env_resonance_type)
    env_preset_table:property("ResonanceEnabled").value =  tostring(resonance_enabled)
    env_preset_table:property("ResonanceLoopType").value =  tostring(res_loop_type)
    env_preset_table:property("ResonanceLoopStart").value =  tostring(res_loop_start)
    env_preset_table:property("ResonanceLoopEnd").value =  tostring(res_loop_end)
    env_preset_table:property("ResonanceSustain").value =  tostring(res_sustain)

    env_preset_table:property("ResonanceLFO").value = tostring(res_lfo_type)
    env_preset_table:property("ResonanceLFOPhase").value = tostring(res_lfo_phase)
    env_preset_table:property("ResonanceLFOFrequency").value = tostring(res_lfo_freq)
    env_preset_table:property("ResonanceLFOAmount").value = tostring(res_lfo_amount)
    env_preset_table:property("ResonanceFollower").value = tostring(res_follow)
    env_preset_table:property("ResonanceFollowerAttack").value = tostring(res_follow_attack)
    env_preset_table:property("ResonanceFollowerRelease").value = tostring(res_follow_release)
    env_preset_table:property("ResonanceFollowerAmount").value = tostring(res_follow_amount)

    env_preset_table:property("CutResFilterType").value =  tostring(cutres_filter_type)


    env_preset_file = preset_root..PATH_SEPARATOR..TOOL_FILE..PATH_SEPARATOR..preset_name..".xml"
    

    check_preset_folder(preset_root)
    if io.exists(env_preset_file) then
      local status = file_exists_dialog(preset_name)
      if status == false then
        return
      end
    end
    local ok,err = env_preset_table:save_as(env_preset_file)    
  
  get_preset_list(vb)  
end


---------------------------------------------------------------------------------------

function load_preset(preset_name,vb)
  --remdebug.engine.start()
  local preset_root = get_tools_root()..TOOL_SAVE_FOLDER..PATH_SEPARATOR..TOOL_FILE
  local instrument = renoise.song().instruments[renoise.song().selected_instrument_index].sample_envelopes
  
    env_preset_file = preset_root..PATH_SEPARATOR..preset_name..".xml"
    get_preset_list(vb)  
    
    local ok,err = env_preset_table:load_from(env_preset_file)  

    preset_version = env_preset_table:property("PresetVersion").value    
    if preset_version == "3.00" then
      renoise.app():show_warning("Sorry, old EA presets are not supported\nLoad in Epic Arpeggiator 3.0,resave...\nThen copy the preset to the preset manager data folder.")
      return
      
    end
    note_scheme_size = MINIMUM_FRAME_LENGTH
    vol_scheme_size = MINIMUM_FRAME_LENGTH
    pan_scheme_size = MINIMUM_FRAME_LENGTH
    note_loop_type = ENV_LOOP_OFF
    vol_loop_type = ENV_LOOP_OFF
    pan_loop_type = ENV_LOOP_OFF    
    envelope_volume_toggle = false
    envelope_panning_toggle = false    

    note_freq_val = 0
    vol_freq_val = 0
    pan_freq_val = 0
    instrument.pitch.play_mode = 1    
    env_preset_name = preset_name
    

      env_pitch_scheme = env_preset_table:property("NoteScheme").value
      note_point_scheme = env_preset_table:property("NotePointSequences").value
      note_scheme_size = tonumber(env_preset_table:property("NoteSchemeSize").value)
      if preset_version == "3.20" then
        envelope_pitch_toggle = toboolean(env_preset_table:property("PitchApply").value)
      else
        envelope_pitch_toggle = true
      end
      
      env_pitch_type = tonumber(env_preset_table:property("NoteDrawType").value)
      note_loop_type = tonumber(env_preset_table:property("NoteLoopType").value)
      note_sustain = tonumber(env_preset_table:property("NoteSustain").value)
      note_loop_start = tonumber(env_preset_table:property("NoteLoopStart").value)
      note_loop_end = tonumber(env_preset_table:property("NoteLoopEnd").value)
      if env_preset_table:property("ToneFactor").value ~= nil then
        local factor_selection = tonumber(env_preset_table:property("ToneFactor").value)
        if factor_selection == ENV_X1 then
            tone_factor = 1
        elseif factor_selection == ENV_X10 then
          tone_factor = 10
        else
          tone_factor = 100
        end

        env_multiplier = tone_factor
      end
      envelope_volume_toggle = toboolean(env_preset_table:property("VolumeApply").value)
      env_vol_scheme = env_preset_table:property("VolumeScheme").value
      volume_point_scheme = env_preset_table:property("VolumePointSequences").value
      vol_scheme_size = tonumber(env_preset_table:property("VolumeSchemeSize").value)
      vol_freq_type = tonumber(env_preset_table:property("VolumeFrequencyMode").value)
      vol_freq_val = tonumber(env_preset_table:property("VolumeFrequency").value)
      env_volume_type = tonumber(env_preset_table:property("VolumeDrawType").value)
      vol_loop_type = tonumber(env_preset_table:property("VolumeLoopType").value)
      vol_loop_start =  tonumber(env_preset_table:property("VolumeLoopStart").value)
      vol_loop_end = tonumber(env_preset_table:property("VolumeLoopEnd").value)
      vol_release = -1
      vol_sustain = -1
      vol_release = tonumber(env_preset_table:property("VolumeRelease").value)
      vol_sustain = tonumber(env_preset_table:property("VolumeSustain").value)
  
      envelope_panning_toggle = toboolean(env_preset_table:property("PanningApply").value)
      env_pan_scheme = env_preset_table:property("PanningScheme").value 
      panning_point_scheme = env_preset_table:property("PanningPointSequences").value 
      pan_scheme_size = tonumber(env_preset_table:property("PanningSchemeSize").value)
      env_panning_type = tonumber(env_preset_table:property("PanningDrawType").value)
      pan_freq_type = tonumber(env_preset_table:property("PanningFrequencyMode").value)
      pan_freq_val = tonumber(env_preset_table:property("PanningFrequency").value)
      pan_loop_type = tonumber(env_preset_table:property("PanningLoopType").value)
      pan_loop_start =  tonumber(env_preset_table:property("PanningLoopStart").value)
      pan_loop_end =  tonumber(env_preset_table:property("PanningLoopEnd").value)
      pan_sustain = -1
      pan_sustain = tonumber(env_preset_table:property("PanningSustain").value)

      note_freq_type = FREQ_TYPE_FREEFORM
      note_freq_val = 0
      vol_freq_type = FREQ_TYPE_FREEFORM
      vol_freq_val = 0
      pan_freq_type = FREQ_TYPE_FREEFORM
      pan_freq_val = 0

   
    if note_loop_end > note_scheme_size then
      note_scheme_size = note_loop_end
      env_note_value[note_loop_end] = NOTE_SCHEME_TERMINATION
    end
    if vol_loop_end > vol_scheme_size then
      vol_scheme_size = vol_loop_end
      env_vol_value[vol_loop_end] = VOL_PAN_TERMINATION
    end
    if pan_loop_end > pan_scheme_size then
      pan_scheme_size = pan_loop_end
      env_pan_value[pan_loop_end] = VOL_PAN_TERMINATION
    end    


    for column = ENV_NOTE_COLUMN, ENV_PAN_COLUMN do
      prepare_note_tables(column)
    end    
    
    init_tables(false)
    
    set_unattended_envelope_values()

    configure_envelope_loop()
    set_pitch_table()
    apply_unattended_properties()        

end

---------------------------------------------------------------------------------------


function set_unattended_envelope_values()
--Values for the LFOs and the cutoff and resonance envelopes
--Only the tables and variables are prepared, nothing is ported to the envelopes yet!
    note_lfo1_type = tonumber(env_preset_table:property("NoteLFO1").value)
    note_lfo1_phase = tonumber(env_preset_table:property("NoteLFOPhase1").value)
    note_lfo1_freq = tonumber(env_preset_table:property("NoteLFOFrequency1").value)
    note_lfo1_amount = tonumber(env_preset_table:property("NoteLFOAmount1").value)
    note_lfo2_type = tonumber(env_preset_table:property("NoteLFO2").value)
    note_lfo2_phase = tonumber(env_preset_table:property("NoteLFOPhase2").value)
    note_lfo2_freq = tonumber(env_preset_table:property("NoteLFOFrequency2").value)
    note_lfo2_amount = tonumber(env_preset_table:property("NoteLFOAmount2").value)
    

    vol_lfo1_type = tonumber(env_preset_table:property("VolumeLFO1").value) 
    vol_lfo1_phase = tonumber(env_preset_table:property("VolumeLFOPhase1").value) 
    vol_lfo1_freq = tonumber(env_preset_table:property("VolumeLFOFrequency1").value) 
    vol_lfo1_amount = tonumber(env_preset_table:property("VolumeLFOAmount1").value) 
    vol_lfo2_type = tonumber(env_preset_table:property("VolumeLFO2").value) 
    vol_lfo2_phase = tonumber(env_preset_table:property("VolumeLFOPhase2").value) 
    vol_lfo2_freq = tonumber(env_preset_table:property("VolumeLFOFrequency2").value) 
    vol_lfo2_amount = tonumber(env_preset_table:property("VolumeLFOAmount2").value) 

    
    pan_lfo1_type = tonumber(env_preset_table:property("PanningLFO1").value) 
    pan_lfo1_phase = tonumber(env_preset_table:property("PanningLFOPhase1").value) 
    pan_lfo1_freq = tonumber(env_preset_table:property("PanningLFOFrequency1").value) 
    pan_lfo1_amount = tonumber(env_preset_table:property("PanningLFOAmount1").value) 
    pan_lfo2_type = tonumber(env_preset_table:property("PanningLFO2").value) 
    pan_lfo2_phase = tonumber(env_preset_table:property("PanningLFOPhase2").value) 
    pan_lfo2_freq = tonumber(env_preset_table:property("PanningLFOFrequency2").value) 
    pan_lfo2_amount = tonumber(env_preset_table:property("PanningLFOAmount2").value)     
    
    
    env_cut_scheme = env_preset_table:property("CutoffScheme").value
    cutoff_point_scheme = env_preset_table:property("CutoffPointSequences").value
    cut_scheme_size = tonumber(env_preset_table:property("CutoffSchemeSize").value) 
    env_cutoff_type = tonumber(env_preset_table:property("CutoffDrawType").value) 
    cutoff_enabled = toboolean(env_preset_table:property("CutoffEnabled").value) 
    cut_loop_type = tonumber(env_preset_table:property("CutoffLoopType").value) 
    cut_loop_start = tonumber(env_preset_table:property("CutoffLoopStart").value) 
    cut_loop_end = tonumber(env_preset_table:property("CutoffLoopEnd").value) 
    cut_sustain = tonumber(env_preset_table:property("CutoffSustain").value) 
    
    cut_lfo_type = tonumber(env_preset_table:property("CutoffLFO").value) 
    cut_lfo_phase = tonumber(env_preset_table:property("CutoffLFOPhase").value) 
    cut_lfo_freq = tonumber(env_preset_table:property("CutoffLFOFrequency").value) 
    cut_lfo_amount = tonumber(env_preset_table:property("CutoffLFOAmount").value) 
    cut_follow = toboolean(env_preset_table:property("CutoffFollower").value) 
    cut_follow_attack = tonumber(env_preset_table:property("CutoffFollowerAttack").value) 
    cut_follow_release = tonumber(env_preset_table:property("CutoffFollowerRelease").value) 
    cut_follow_amount = tonumber(env_preset_table:property("CutoffFollowerAmount").value) 


    env_res_scheme = env_preset_table:property("ResonanceScheme").value
    resonance_point_scheme = env_preset_table:property("ResonancePointSequences").value
    res_scheme_size = tonumber(env_preset_table:property("ResonanceSchemeSize").value) 
    env_resonance_type = tonumber(env_preset_table:property("ResonanceDrawType").value) 
    resonance_enabled = toboolean(env_preset_table:property("ResonanceEnabled").value) 
    res_loop_type = tonumber(env_preset_table:property("ResonanceLoopType").value) 
    res_loop_start = tonumber(env_preset_table:property("ResonanceLoopStart").value) 
    res_loop_end = tonumber(env_preset_table:property("ResonanceLoopEnd").value) 
    res_sustain = tonumber(env_preset_table:property("ResonanceSustain").value) 

    res_lfo_type = tonumber(env_preset_table:property("ResonanceLFO").value) 
    res_lfo_phase = tonumber(env_preset_table:property("ResonanceLFOPhase").value) 
    res_lfo_freq = tonumber(env_preset_table:property("ResonanceLFOFrequency").value) 
    res_lfo_amount = tonumber(env_preset_table:property("ResonanceLFOAmount").value) 
    res_follow = toboolean(env_preset_table:property("ResonanceFollower").value) 
    res_follow_attack = tonumber(env_preset_table:property("ResonanceFollowerAttack").value) 
    res_follow_release = tonumber(env_preset_table:property("ResonanceFollowerRelease").value) 
    res_follow_amount = tonumber(env_preset_table:property("ResonanceFollowerAmount").value) 

    cutres_filter_type = tonumber(env_preset_table:property("CutResFilterType").value) 

    populate_cutres_tables(ENV_CUT)    
    populate_cutres_tables(ENV_RES)    
end


---------------------------------------------------------------------------------------

function prepare_note_tables(column)
--Convert the strings to tables to show in the tool-track
  local preset_version = env_preset_table:property("PresetVersion").value
  local freq_type = FREQ_TYPE_FREEFORM
  local freq_val = 0
  if column == ENV_NOTE_COLUMN then
--  print('initialising tables')
    init_tables(ENV_NOTE_COLUMN)    
    
  elseif column == ENV_VOL_COLUMN then
    init_tables(ENV_VOL_COLUMN)    
    
  elseif column == ENV_PAN_COLUMN then
    init_tables(ENV_PAN_COLUMN)  
  end

--    print('freeform')
    local y_value_table = {}
    local x_value_table = {}
    local end_node = #y_value_table
    local leading_table = 1 --Notes
    local sub_position = 1

    if column == ENV_NOTE_COLUMN then
      y_value_table = env_pitch_scheme:split( "[^,%s]+" )
      x_value_table = note_point_scheme:split( "[^,%s]+" )
      note_freq_val = 0
    elseif column == ENV_VOL_COLUMN and tonumber(preset_version) >= 3.10 then
      y_value_table = env_vol_scheme:split( "[^,%s]+" )
      x_value_table = volume_point_scheme:split( "[^,%s]+" )
      vol_freq_val = 0
    elseif column == ENV_PAN_COLUMN and tonumber(preset_version) >= 3.10 then
      y_value_table = env_pan_scheme:split( "[^,%s]+" )
      x_value_table = panning_point_scheme:split( "[^,%s]+" )
      pan_freq_val = 0
    end
          
    end_node = #y_value_table
          
    if #y_value_table < #x_value_table then
      end_node = #x_value_table
      leading_table = 2
    end
          
    if #y_value_table == #x_value_table then
      leading_table = -1
    end
          
    if leading_table == 1 then
      sub_position = 2
            
      for t = #x_value_table+1, end_node do
        local new_val = tonumber(x_value_table[sub_position]) -
                        tonumber(x_value_table[sub_position -1])
        x_value_table[t] = tonumber(x_value_table[t-1])+new_val
        sub_position = sub_position + 1
      end
    else
      for t = #y_value_table+1, end_node do
        y_value_table[t] = y_value_table[sub_position]
        sub_position = sub_position + 1
        if sub_position > #y_value_table then
          sub_position = 1
        end
      end
    end
    
    for t = 1, end_node do
      
      if y_value_table[t] ~= nil and x_value_table[t] ~= nil then
        if column == ENV_NOTE_COLUMN then

          env_note_value[tonumber(x_value_table[t])] = y_value_table[t]
          
          if t == end_node then
            note_scheme_size = tonumber(x_value_table[t])
          end
        elseif column == ENV_VOL_COLUMN and tonumber(preset_version) >= 3.10 then
          env_vol_value[tonumber(x_value_table[t])] = y_value_table[t]
          if t == end_node then
            vol_scheme_size = tonumber(x_value_table[t])
          end
        elseif column == ENV_PAN_COLUMN and tonumber(preset_version) >= 3.10 then
          env_pan_value[tonumber(x_value_table[t])] = y_value_table[t]
          if t == end_node then
            pan_scheme_size = tonumber(x_value_table[t])
          end
        end
      end
    end

    for t = 0, note_scheme_size-1 do
      if env_note_value[t] == NOTE_SCHEME_TERMINATION then
        env_note_value[t] = EMPTY_CELL
      end
    end
    for t = 0, vol_scheme_size-1 do
      if  env_vol_value[t] == VOL_PAN_TERMINATION then
        env_vol_value[t] = EMPTY_CELL
      end
    end
    for t = 0, pan_scheme_size-1 do
      if  env_vol_value[t] == VOL_PAN_TERMINATION then
        env_vol_value[t] = EMPTY_CELL
      end
    end
  
  if note_loop_type ~= ENV_LOOP_OFF or auto_note_loop == ARP_MODE_AUTO then
    note_loop_start = 0
    note_loop_end = note_scheme_size
  else
    note_loop_start = -1
    note_loop_end = -1
  end
  if tonumber(env_preset_table:property("PresetVersion").value) >= 3.10 then
    if vol_loop_type ~= ENV_LOOP_OFF or auto_vol_loop == ARP_MODE_AUTO then
      vol_loop_start = 0
      vol_loop_end = vol_scheme_size
    else
      vol_loop_start = -1
      vol_loop_end = -1
    end
    if pan_loop_type ~= ENV_LOOP_OFF or auto_pan_loop == ARP_MODE_AUTO then
      pan_loop_start = 0
      pan_loop_end = pan_scheme_size
    else
      pan_loop_start = -1
      pan_loop_end = -1
    end
  end

end


---------------------------------------------------------------------------------------

function show_folder()
  local preset_root = get_tools_root()..TOOL_SAVE_FOLDER
  local preset_folder = preset_root..PATH_SEPARATOR..TOOL_FILE
  check_preset_folder(preset_root)
  renoise.app():open_path(preset_folder)
end


---------------------------------------------------------------------------------------

function check_preset_folder(preset_root)

  local tool_folders = os.dirnames(get_tools_root())
  
  local env_preset_folder_present = false
  local save_data_folder_present = false
  
  for _ = 1, #tool_folders do
    if tool_folders[_] == TOOL_SAVE_FOLDER then
      save_data_folder_present = true
      break
    end
  end
  if save_data_folder_present == false then
    os.mkdir(get_tools_root()..TOOL_SAVE_FOLDER)
  end

  local root_folders = os.dirnames(preset_root)
  
  for _ = 1, #root_folders do
    if root_folders[_] == TOOL_FILE then
      env_preset_folder_present = true
      break
    end
  end

  if env_preset_folder_present == false then
    os.mkdir(preset_root..PATH_SEPARATOR..TOOL_FILE)
  end

end



--------------------------------------------------------------------------------
--      End of teh road...
--------------------------------------------------------------------------------

